import sys
import os

cpu_index = 0
cpu_total = 1
job_index = 0

if len(sys.argv)>=3:
	cpu_index = int(sys.argv[1])
	cpu_total = int(sys.argv[2])
else:
	print("warning : arguments might be missing")

def add_job(func):
	global job_index
	if job_index == cpu_index:
		print(f"I'm running job {job_index}")
		func()
	job_index = (job_index+1)%cpu_total

###############################################################

from functools import reduce # Valid in Python 2.6+, required in Python 3
import operator

import torch
import torch.nn as nn
import torch.nn.functional as F

import anyact as aa

import numpy as np
np.seterr(divide='ignore', invalid='ignore')

import torchvision
import torchvision.datasets as datasets

import os
import h5py

from toy_problem_mnist import fmnist,mnist,mnist_perf

import matplotlib.pyplot as plt

fmnist_L_x,fmnist_L_y = fmnist(True)
fmnist_T_x,fmnist_T_y = fmnist(False)

def lerp(a,b,λ): return λ*b + (1-λ)*a

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

def nn_fc(layers,σ,dropout=0.5,batchnorm=1,gen_linear=lambda dim_i,dim_o: nn.Linear(dim_i,dim_o)):
	layers = [int(np.round(x)) for x in layers]
	disable_drp = 0 if dropout==0 else 1
	disable_btn = batchnorm
	if hasattr(σ,"m"): m=σ.m
	else: m = 1
	return sum([
		[gen_linear((m if not index==0 else 1)*λ1,λ2),σ,
		]+disable_btn*[nn.BatchNorm1d(λ2*m),
		]+disable_drp*[aa.Dropout(dropout),
		]
		for index,(λ1,λ2) in enumerate(zip(layers[0:-2],layers[1:-1]))
	],[])+[gen_linear(m*layers[-2],layers[-1])] # remove last GELU, BatchNorm and Dropout

# x=torch.randn(5,1,28,28)
# nn.Sequential(nn.Flatten(),*nn_fc([28*28,30,10],aa.Stack([nn.GELU(),nn.GELU()]),dropout=0.5,batchnorm=1))(x).shape

from toy_problem_mnist_analysis import analysis_classification

def expspace(x0,x1,N): return torch.exp(torch.linspace(np.log(x0),np.log(x1),N))

from gradiator import gradiator_1

def cantor():
	if not hasattr(cantor, 'array'): cantor.array = aa.cantor_function(torch.linspace(0,1,3**5+1),5).to(device)
	return aa.modulize(gradiator_1(
		lambda x: torch.floor(x) + aa.sampler1d_repeat(cantor.array,x,aa.nearest_floor),
		# lambda x: aa.sampler1d_clamp(cantor.array,x,aa.nearest_floor),
		lambda x,y,grad: torch.zeros(x.shape,device=x.device) # derivative is 0 everywhere
	))

def cantor_fakegrad():
	if not hasattr(cantor, 'array'): cantor.array = aa.cantor_function(torch.linspace(0,1,3**5+1),5).to(device)
	return aa.modulize(gradiator_1(
		lambda x: torch.floor(x) + aa.sampler1d_repeat(cantor.array,x,aa.nearest_floor),
		lambda x,y,grad: grad # fake the gradient to identity
	))

def square_fakegrad():
	return aa.modulize(gradiator_1(
		lambda x: torch.round(torch.sin(x)),
		lambda x,y,grad: torch.cos(x)*grad # gradient forgets about rounding
	))

def pshitLU():
	return aa.modulize(gradiator_1(
		lambda x: torch.relu(x + .5*torch.randn(x.shape)) + .1*torch.randn(x.shape),
		lambda x,y,grad: (x>0)*grad # gradient forgets about noise
	))

def pshitLU2():
	return aa.modulize(gradiator_1(
		lambda x: torch.relu(x),
		lambda x,y,grad: (x+0.5*torch.randn(x.shape)>0)*grad # gradient adds noise
	))

def debuuuug(act):
	a = nn.Linear(1,1).to(device)
	optim = torch.optim.Adam(a.parameters(),lr=1e-2)
	x = torch.ones(1).to(device)
	for i in range(128):
		loss = torch.abs(1-act(a(x)))
		optim.zero_grad()
		loss.backward()
		optim.step()
		print(loss.item())

# debuuuug(cantor())
# debuuuug(cantor_fakegrad())
# debuuuug(square_fakegrad())

# pshitLU()
# pshitLU2()

def perlin():
	print("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA LOST THAT CODE")

class ParametricSin(nn.Module):
	def __init__(self,x_shape=[1]):
		super().__init__()
		self.θ = nn.parameter.Parameter(torch.ones(x_shape))
	def forward(self,x):
		return torch.sin(self.θ*x)

# x = torch.randn(10,4)
# psin = ParametricSin(x.shape)
# psin(x)

# for loss_name,loss in [["MSE",nn.MSELoss()],]:
# 	for σ_name,σ in [["exp_other_lr",aa.modulize(torch.sin)]]:
# 		for N in [2,4,16]:
# 			for dropout in [0,0.5]:
# 				for batchnorm in [1,0]:
# 					add_job(lambda: analysis_classification(
# 						nn.Sequential(
# 							nn.Flatten(),
# 							*nn_fc(expspace(28*28,10,N+1),σ,dropout=dropout,batchnorm=batchnorm),
# 							nn.Softmax(dim=1)
# 						),
# 						f"fc({N})-{σ_name}-dropout={dropout}-batchnorm={batchnorm}",
# 						loss,
# 						lr0=-4,lr1=-6,epochs=256
# 					))


# for loss_name,loss in [["MSE",nn.MSELoss()],]:
# 	for σ_name,σ in [["sin_other_lr",aa.modulize(torch.sin)]]:
# 		for N in [2,4,16]:
# 			for dropout in [0,0.5]:
# 				for batchnorm in [1,0]:
# 					add_job(lambda: analysis_classification(
# 						nn.Sequential(
# 							nn.Flatten(),
# 							*nn_fc(expspace(28*28,10,N+1),σ,dropout=dropout,batchnorm=batchnorm),
# 							nn.Softmax(dim=1)
# 						),
# 						f"fc({N})-{σ_name}-dropout={dropout}-batchnorm={batchnorm}",
# 						loss,
# 						lr0=-2,lr1=-6,epochs=256
# 					))

# x = torch.linspace(-3,3,3**10+1)
# x = torch.linspace(-2,2,3**10+1)
# plt.plot(x,cantor_fakegrad()(x.to('cuda')).cpu().numpy())
# plt.plot(x,cantor()(x.to('cuda')).cpu().numpy())
# plt.plot(x,aa.sampler1d_repeat(cantor.array,x.to(device),aa.nearest).cpu().numpy())

########################
### TOTALEMENT CON ! ###
########################
# def gradiator_1_noise(noise,fun,grad_fun):
# 	class GradiatorFunction(torch.autograd.Function):
# 		@staticmethod
# 		def forward(ctx,A,x):
# 			x = x + noise*torch.randn(x.shape,device=x.device)
# 			y = fun(x)
# 			ctx.save_for_backward(x,y) # save input for backward pass
# 			return y
# 		@staticmethod
# 		def backward(ctx, grad_out):
# 			# if not ctx.needs_input_grad[0]: return None # if grad not required, don't compute
# 			x,y = ctx.saved_tensors # restore input from context
# 			grad_in = grad_fun(x,y,grad_out)
# 			return grad_in
# 	return GradiatorFunction.apply

def gradiator_noise(noise,fun,grad_fun):
	class GradiatorFunction(torch.autograd.Function):
		@staticmethod
		def forward(ctx,A,x):
			x = x + noise*torch.randn(x.shape,device=x.device)
			y = fun(A,x)
			ctx.save_for_backward(A,x,y) # save input for backward pass
			return y
		@staticmethod
		def backward(ctx, grad_out):
			# if not ctx.needs_input_grad[0]: return None # if grad not required, don't compute
			A,x,y = ctx.saved_tensors # restore input from context
			grad_in = grad_fun(A,x,y,grad_out)
			return grad_in
	return GradiatorFunction.apply

# class LinearNoise(nn.Module):
# 	def __init__(self,dim_i,dim_o):
# 		super().__init__()
# 		self.A = nn.parameter.Parameter(torch.randn(dim_i,dim_o))
# 		def fwd(A,x):
# 			return torch.einsum('ij,ni->nj', A, x)
# 		def bwd(A,x,y,grad_out):
# 			grad_in_A = torch.einsum('ni,nj->nij', x, grad_out) # produit extérieur
# 			grad_in_x = torch.einsum('nj,ij->ni' , grad_out, A) # A⋅x
# 			return grad_in_A, grad_in_x
# 		self.model = gradiator_noise(0.5,fwd,bwd)
# 	def forward(self,x):
# 		return self.model(self.A,x)

# # for N in [2,4,8]:
# for N in [4,8]:
# 	analysis_classification(
# 		nn.Sequential(
# 			nn.Flatten(),
# 			*nn_fc(expspace(28*28,10,N+1),nn.GELU(),dropout=0,batchnorm=1,gen_linear=lambda dim_i,dim_o: LinearNoise(dim_i,dim_o)),
# 			nn.Softmax(dim=1)
# 		),
# 		f"fc({N})-LinearNoise",
# 		nn.MSELoss(),
# 	)

def gradiator_1_noise(noise,fun,grad_fun):
	class GradiatorFunction(torch.autograd.Function):
		@staticmethod
		def forward(ctx,A,x):
			x = x + noise*torch.randn(x.shape,device=x.device)
			y = fun(x)
			ctx.save_for_backward(x,y) # save input for backward pass
			return y
		@staticmethod
		def backward(ctx, grad_out):
			# if not ctx.needs_input_grad[0]: return None # if grad not required, don't compute
			x,y = ctx.saved_tensors # restore input from context
			grad_in = grad_fun(x,y,grad_out)
			return grad_in
	return GradiatorFunction.apply

class GELUNoise(nn.Module):
	def __init__(self,noise):
		super().__init__()
		self.noise = noise
	def forward(self,x):
		return F.gelu(x+self.noise*torch.randn(x.shape,device=x.device))

noise = 5

x = torch.linspace(-30,30,2**17)
plt.clf()
plt.scatter(x,GELUNoise(noise)(x),s=1,alpha=1/64)
plt.savefig(f"out/GELUNoise{noise}-preview.png")

# for N in [2,4,8]:
# for N in [4,8]:
for N in [8]:
	analysis_classification(
		nn.Sequential(
			nn.Flatten(),
			*nn_fc(expspace(28*28,10,N+1),GELUNoise(noise),dropout=0,batchnorm=1),
			nn.Softmax(dim=1)
		),
		f"fc({N})-GELUNoise{noise}",
		nn.MSELoss(),
	)

###

class SinNoise(nn.Module):
	def __init__(self,noise):
		super().__init__()
		self.noise = noise
	def forward(self,x):
		return torch.sin(x+self.noise*torch.randn(x.shape,device=x.device))

for noise in [.5,1]:
	x = torch.linspace(-6,6,2**17)
	plt.clf()
	plt.scatter(x,SinNoise(noise)(x),s=1,alpha=1/64)
	plt.savefig(f"out/SinNoise{noise}-preview.png")
	# for N in [2,4,8]:
	# for N in [2]:
	# for N in [4]:
	# for N in [8]:
		analysis_classification(
			nn.Sequential(
				nn.Flatten(),
				*nn_fc(expspace(28*28,10,N+1),SinNoise(noise),dropout=0,batchnorm=1),
				nn.Softmax(dim=1)
			),
			f"fc({N})-SinNoise{noise}",
			nn.MSELoss(),
		)





#############################################################################
#############################################################################
#############################################################################
#############################################################################


class FixedNoise(nn.Module):
	def __init__(self):
		super().__init__()
		self.content=
	def forward(self,x):
		return aa.sampler1d_repeat(tex,x,interpolation=lerp)

x2_bn1 = nn.Sequential(
	nn.Flatten(),
	*nn_fc(expspace(28*28,10,N+1),aa.modulize(lambda x:x**2),dropout=0,batchnorm=1),
	nn.Softmax(dim=1)
)

x2_bn0 = nn.Sequential(
	nn.Flatten(),
	*nn_fc(expspace(28*28,10,N+1),aa.modulize(lambda x:x**2),dropout=0,batchnorm=0),
	nn.Softmax(dim=1)
)

analysis_classification(x2_bn1, f"fc({N})-x2-batchnorm={1}", nn.MSELoss(),)
analysis_classification(x2_bn0, f"fc({N})-x2-batchnorm={0}", nn.MSELoss(),)

plt.clf()
plt.hist(torch.concat([torch.flatten(p.detach().cpu()) for p in x2_bn1.parameters()]),512,density=True)
plt.savefig(f"out/fc({N})-x2-batchnorm={1}-param-hist")

plt.clf()
plt.hist(torch.concat([torch.flatten(p.detach().cpu()) for p in x2_bn0.parameters()]),512,density=True)
plt.savefig(f"out/fc({N})-x2-batchnorm={0}-param-hist")

#############################################################################
#############################################################################
#############################################################################
#############################################################################














class Jeff(nn.Module):
	def __init__(self):
		super().__init__()
	def forward(self,x):
		x.view(x.shape[0],-1,2)

analysis_classification(
	nn.Sequential(
		nn.Flatten(),
		*nn_fc(expspace(28*28,10,N+1),Jeff(),dropout=0,batchnorm=0),
		nn.Softmax(dim=1)
	),
	f"fc({N})-SinNoise{noise}",
	nn.MSELoss(),
)




############## MAIN STUFF ##############

σs=[
	### https://www.youtube.com/watch?v=Ae9EKCyI1xU
	["Identity",nn.Identity()],
	### RELU LIKE
	["GELU"          ,nn.GELU()],
	["ReLU"          ,nn.ReLU()],
	# ["ReLU6"         ,nn.ReLU6()],
	["RReLU"         ,nn.RReLU(inplace=True)],
	["PReLU"         ,nn.PReLU()],
	["SiLU"          ,nn.SiLU()],
	["ELU"           ,nn.ELU(inplace=True)],
	["SELU"          ,nn.SELU(inplace=True)],
	["GELU"          ,nn.GELU()],
	["CELU"          ,nn.CELU()],
	["Mish"          ,nn.Mish(inplace=True)],
	["Softplus"      ,nn.Softplus()],
	["Hardswish"     ,nn.Hardswish(inplace=True)],
	["LeakyReLU"     ,nn.LeakyReLU(inplace=True)],
	["LeakyReLU(0.1)",nn.LeakyReLU(0.1,inplace=True)],
	["LeakyReLU(0.5)",nn.LeakyReLU(0.5,inplace=True)],
	["LeakyReLU(5.0)",nn.LeakyReLU(5.0,inplace=True)],
	# ["Threshold"  ,nn.Threshold(inplace=True)],
	### CLASSICAL
	["Softsign"   ,nn.Softsign()],
	["Tanh"       ,nn.Tanh()],
	["Hardtanh"   ,nn.Hardtanh(inplace=True)],
	["sigmoid"    ,nn.Sigmoid()],
	["Hardsigmoid",nn.Hardsigmoid(inplace=True)],
	### OTHER - RELU-like but in both direction
	["Softshrink" ,nn.Softshrink()],
	["Hardshrink" ,nn.Hardshrink()],
	["Tanhshrink" ,nn.Tanhshrink()],
	["abs"        ,aa.modulize(torch.abs)],
	### TRIGO
	["atan"       ,aa.modulize(torch.atan)],
	["tan"        ,aa.modulize(torch.tan)],
	["sinc"       ,aa.modulize(torch.sinc)],
	["sin"        ,aa.modulize(torch.sin)],
	["sin(πx)"    ,aa.modulize(lambda x:torch.sin(np.pi*x))],
	### STACKING
	["ExpLog"     ,aa.Stack([aa.Log(),aa.Exp()])], # TODO 
	["IdGELU"     ,aa.Stack([nn.Identity(),nn.GELU()])],
	["IdGELUsin"  ,aa.Stack([nn.Identity(),nn.GELU(),aa.modulize(torch.sin)])],
	["Idsin"      ,aa.Stack([nn.Identity(),aa.modulize(torch.sin)])],
	### PROBABLY INAPPROPRIATE
	["Softmax"    ,nn.Softmax()],
	["Softmin"    ,nn.Softmin()],
	["LogSigmoid" ,nn.LogSigmoid()],
	["LogSoftmax" ,nn.LogSoftmax()],
	### MESSED UP
	["exp"        ,aa.modulize(torch.exp)],
	# ["log"        ,aa.modulize(torch.log)],
	["log_abs"    ,aa.modulize(lambda x:torch.log(1e-5+torch.abs(x)))],
	["sqrt"       ,aa.modulize(torch.sqrt)],
	["x^2"        ,aa.modulize(lambda x: x*x)],
	### NOT DIFFERENTIABLE
	["cantor"     ,cantor()],
	["cantor"     ,cantor_fakegrad()],
	["square_fakegrad"     ,square_fakegrad()],
	["perlin"     ,perlin()],
	# TODO : perlin noise
	# TODO : GLSL noise function
	# TODO : jeff
	# TODO : aa.Grid(σ1,σ2,σ3...)
	# TODO : aa.Alternate(σ1,σ2,σ3...)
]
# print(len(σs))
# x = torch.randn(20,1,28,28)
# for σ_name,σ in σs:
# 	try:
# 		print("######################")
# 		# print(σ_name,nn_fc(expspace(28*28,10,5+1),σ,dropout=.5,batchnorm=1))
# 		print(
# 			nn.Sequential(
# 				nn.Flatten(),
# 				*nn_fc(expspace(28*28,10,1+1),σ,dropout=.5,batchnorm=0),
# 				nn.Softmax(dim=1)
# 			)(x).shape
# 			,σ_name
# 		)
# 	except:
# 		print("!!!!!!!!!!!!!!!!!!!!!!!", σ_name)

# nn.Sequential(
# 	nn.Flatten(),
# 	*nn_fc(expspace(28*28,10,5+1),aa.Stack([aa.Exp(),aa.Log()]),dropout=.5,batchnorm=0),
# 	nn.Softmax(dim=1)
# )(x).shape


# σs = [
# 	["cantor2"    ,cantor2()],
# 	["cantor"     ,cantor()],
# 	# ["sinc(2x)"   ,modulize(lambda x: torch.sinc(2*x))],
# 	# ["fourier_1",modulize(lambda x: sum([A*torch.sin(2*π*ω*x+φ) for A,ω,φ in [ [0,0,π],[1,1,0],[1,2,0],[1,3,0], ]]))],
# 	# ["fourier_2",modulize(lambda x: sum([A*torch.sin(2*π*ω*x+φ) for A,ω,φ in [ [0,0,π],[1,1,π],[1,2,π/8],[-.2,3,π/4], ]]))],
# ]

# https://humphryscomputing.com/Notes/Neural/chaos.html
	# ["log(exp(-1)+abs(x))"        ,modulize(lambda x: torch.log(math.exp(-1)+torch.abs(x)))],
	# ["log(abs(x))"        ,modulize(lambda x: torch.log(torch.abs(x)))],
	# ["exp(x)"          ,modulize(lambda x: torch.exp(x))],
	# ["exp(x)+log(abs(x))" ,modulize(lambda x: torch.exp(x)+torch.log(torch.abs(x)))],
	# ["exp(sin(x)⁻¹)"  , modulize(lambda x: torch.exp(1/torch.sin(x)))],
	# ["exp(sin(πx)⁻¹)" , modulize(lambda x: torch.exp(1/torch.sin(π*x)))],
	# ["exp(sin(2πx)⁻¹)", modulize(lambda x: torch.exp(1/torch.sin(2*π*x)))],
	# ["exp(sin(3πx)⁻¹)", modulize(lambda x: torch.exp(1/torch.sin(3*π*x)))],
	# ["exp(x)-1"                   ,modulize(lambda x: torch.exp(x)-1)],
	# ["exp(x)-1+log(exp(-1)+abs(x))" ,modulize(lambda x: torch.exp(x)-1+torch.log(math.exp(-1)+torch.abs(x)))],
	# ["-exp(x)+1-log(exp(-1)+abs(x))" ,modulize(lambda x:-torch.exp(x)+1-torch.log(math.exp(-1)+torch.abs(x)))],
	# ["exp(sin(x)⁻¹)-1"  , modulize(lambda x: torch.exp(1/(1+torch.sin(    x))-1))],
	# ["exp(sin(1πx)⁻¹)-1", modulize(lambda x: torch.exp(1/(1+torch.sin(1*π*x))-1))],
	# ["exp(sin(2πx)⁻¹)-1", modulize(lambda x: torch.exp(1/(1+torch.sin(2*π*x))-1))],
	# ["exp(sin(3πx)⁻¹)-1", modulize(lambda x: torch.exp(1/(1+torch.sin(3*π*x))-1))],
	# # # ["sin-mess-a",modulize(lambda x: (2*torch.sin(3/x))+(3*torch.cos(5/x))+(4*torch.sin(6/x))+(1*torch.cos(3/x)))],
	# # # ["sin-mess-b",modulize(lambda x: torch.sin(3/x)*torch.sin(5/(1-x)))],
	# # # ["sin-mess-b",modulize(lambda x: torch.sin(3*x)*torch.sin(5/(1-x)))],
	# # # ["sin-mess-b",modulize(lambda x: torch.sin(2/(1-x)))],
	# # # ["sin-mess-b",modulize(lambda x: torch.sin(2*π/(.5-x**2)))],
	# # # ["sin-mess-b",modulize(lambda x: torch.sin(5/(1-x)))],
	# ["cos(8πx)⋅cos(4πx)",modulize(lambda x: torch.cos(8*π*x)*torch.cos(4*π*x))],
	# ["cos(8πx)⋅cos(4πx)⋅cos(2πx)",modulize(lambda x: torch.cos(8*π*x)*torch.cos(4*π*x)*torch.cos(2*π*x))],
	# ["cos((x-0.0)⁻¹)",modulize(lambda x: torch.cos(1/x))],
	# ["cos((x-0.5)⁻¹)",modulize(lambda x: torch.cos(1/(x-.5)))],
	# ["cos((x-1.0)⁻¹)",modulize(lambda x: torch.cos(1/(x-1)))],
	# ["cos((x-2.0)⁻¹)",modulize(lambda x: torch.cos(1/(x-2)))],
	# ["cos(5(x-2.0)⁻¹)",modulize(lambda x: torch.cos(5/(x-2)))],
	# ["crazy-cos",modulize(lambda x: torch.cos(5/(0.2+1+torch.cos(π*x))))],
	# # ["sin(8πx)⋅sin(4πx)",modulize(lambda x: torch.sin(8*π*x)*torch.sin(4*π*x))],
	# # ["sin(8πx)⋅sin(4πx)⋅sin(2πx)",modulize(lambda x: torch.sin(8*π*x)*torch.sin(4*π*x)*torch.sin(2*π*x))],
	# ["sin((x-0.0)⁻¹)",modulize(lambda x: torch.sin(1/x))],
	# # ["sin((x-0.5)⁻¹)",modulize(lambda x: torch.sin(1/(x-.5)))],
	# ["sin((x-1.0)⁻¹)",modulize(lambda x: torch.sin(1/(x-1)))],
	# # ["sin((x-2.0)⁻¹)",modulize(lambda x: torch.sin(1/(x-2)))],
	# # ["sin(5(x-2.0)⁻¹)",modulize(lambda x: torch.sin(5/(x-2)))],
	# ["cos(πx)⁻¹"  , modulize(lambda x: 1./(torch.cos(π*x)))],
	# ["sin(cos(πx)⁻¹)"  , modulize(lambda x: torch.sin(1./(torch.cos(π*x))))],
	# ["sin(cos(πx⋅2⁻¹)⁻¹)"  , modulize(lambda x: torch.sin(1./(torch.cos(π*x/2))))],


# for loss_name,loss in [["MSE",nn.MSELoss()],]:
# 	for σ_name,σ in σs:
# 		for N in [2,4,16]:
# 			for dropout in [0,0.5]:
# 				for batchnorm in [1,0]:
# 					add_job(lambda: analysis_classification(
# 						nn.Sequential(
# 							nn.Flatten(),
# 							*nn_fc(expspace(28*28,10,N+1),σ,dropout=dropout,batchnorm=batchnorm),
# 							nn.Softmax(dim=1)
# 						),
# 						f"fc({N})-{σ_name}-dropout={dropout}-batchnorm={batchnorm}",
# 						loss
# 					))

# def nn_cn(N,m_size,M,ker_size,σ):
# 	return nn.Sequential(
# 		nn.Conv2d(), nn.BatchNorm2d(), 
# 		nn.Flatten(),
# 		*nn_fc(torch.linspace(28*28,10,N+1),σ,dropout=dropout),
# 		nn.Softmax(dim=1)
# 	),

# def nn_cn(channels,ker_size,σ,m_size=1,dropout=0):
# 	pad1a = (ker_size-1)//2
# 	pad1b = (ker_size  )//2
# 	pad2a = (  m_size-1)//2
# 	pad2b = (  m_size  )//2
# 	channels = [int(c) for c in channels]
# 	return [
# 			nn.Conv2d(λ1,λ2,ker_size)      if index==0 else
# 			Pad((pad1a,pad1b,pad1a,pad1b)) if index==1 else
# 			nn.BatchNorm2d(λ2)             if index==2 else
# 			σ                              if index==3 else
# 			nn.MaxPool2d(m_size,stride=1)  if index==4 else
# 			Pad((pad2a,pad2b,pad2a,pad2b)) if index==5 else
# 			nn.Dropout(dropout)
# 			for λ1,λ2 in zip(channels[:-1],channels[1:])
# 			for index in range(7)
# 		]


# for m_size in range(1,10):
# 	nn.Sequential(*nn_cn([1,64,64],ker_size=3,m_size=m_size,σ=nn.GELU(),dropout=0))(torch.randn(10,1,28,28)).shape




